## -*- coding: utf-8 -*-
# test_events.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import unittest
import sets
import time
import socket
import threading
import random


from mock import Mock
from protobuf.socketrpc import RpcService
from leap.common import events
from leap.common.events import (
    server,
    client,
    mac_auth,
)
from leap.common.events.events_pb2 import (
    EventsServerService,
    EventsServerService_Stub,
    EventsClientService_Stub,
    EventResponse,
    SignalRequest,
    RegisterRequest,
    PingRequest,
    SOLEDAD_CREATING_KEYS,
    CLIENT_UID,
)


port = 8090

received = False


class EventsTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        server.EventsServerDaemon.ensure(8090)
        cls.callbacks = events.client.registered_callbacks

    @classmethod
    def tearDownClass(cls):
        # give some time for requests to be processed.
        time.sleep(1)

    def setUp(self):
        super(EventsTestCase, self).setUp()

    def tearDown(self):
        #events.client.registered_callbacks = {}
        server.registered_clients = {}
        super(EventsTestCase, self).tearDown()

    def test_service_singleton(self):
        """
        Ensure that there's always just one instance of the server daemon
        running.
        """
        service1 = server.EventsServerDaemon.ensure(8090)
        service2 = server.EventsServerDaemon.ensure(8090)
        self.assertEqual(service1, service2,
                         "Can't get singleton class for service.")

    def test_client_register(self):
        """
        Ensure clients can register callbacks.
        """
        self.assertTrue(1 not in self.callbacks,
                        'There should should be no callback for this signal.')
        events.register(1, lambda x: True)
        self.assertTrue(1 in self.callbacks,
                        'Could not register signal in local client.')
        events.register(2, lambda x: True)
        self.assertTrue(1 in self.callbacks,
                        'Could not register signal in local client.')
        self.assertTrue(2 in self.callbacks,
                        'Could not register signal in local client.')

    def test_register_signal_replace(self):
        """
        Make sure clients can replace already registered callbacks.
        """
        sig = 3
        cbk = lambda x: True
        events.register(sig, cbk, uid=1)
        self.assertRaises(Exception, events.register, sig, lambda x: True,
                          uid=1)
        events.register(sig, lambda x: True, uid=1, replace=True)
        self.assertTrue(sig in self.callbacks, 'Could not register signal.')
        self.assertEqual(1, len(self.callbacks[sig]),
                         'Wrong number of registered callbacks.')

    def test_signal_response_status(self):
        """
        Ensure there's an appropriate response from server when signaling.
        """
        sig = 4
        request = SignalRequest()
        request.event = sig
        request.content = 'my signal contents'
        request.mac_method = mac_auth.MacMethod.MAC_NONE
        request.mac = ""
        service = RpcService(EventsServerService_Stub, port, 'localhost')
        # test synch
        response = service.signal(request, timeout=1000)
        self.assertEqual(EventResponse.OK, response.status,
                         'Wrong response status.')

    def test_signal_executes_callback(self):
        """
        Ensure callback is executed upon receiving signal.
        """
        sig = CLIENT_UID
        request = SignalRequest()
        request.event = sig
        request.content = 'my signal contents'
        request.mac_method = mac_auth.MacMethod.MAC_NONE
        request.mac = ""
        service = RpcService(EventsServerService_Stub, port, 'localhost')

        # register a callback
        flag = Mock()
        events.register(sig, lambda req: flag(req.event))
        # signal
        response = service.signal(request)
        self.assertEqual(EventResponse.OK, response.status,
                         'Wrong response status.')
        time.sleep(1)  # wait for signal to arrive
        flag.assert_called_once_with(sig)

    def test_events_server_service_register(self):
        """
        Ensure the server can register clients to be signaled.
        """
        sig = 5
        request = RegisterRequest()
        request.event = sig
        request.port = 8091
        request.mac_method = mac_auth.MacMethod.MAC_NONE
        request.mac = ""
        service = RpcService(EventsServerService_Stub, port, 'localhost')
        complist = server.registered_clients
        self.assertEqual({}, complist,
                         'There should be no registered_ports when '
                         'server has just been created.')
        response = service.register(request, timeout=1000)
        self.assertTrue(sig in complist, "Signal not registered succesfully.")
        self.assertTrue(8091 in complist[sig],
                        'Failed registering client port.')

    def test_client_request_register(self):
        """
        Ensure clients can register themselves with server.
        """
        sig = 6
        complist = server.registered_clients
        self.assertTrue(sig not in complist,
                        'There should be no registered clients for this '
                        'signal.')
        events.register(sig, lambda x: True)
        time.sleep(0.1)
        port = client.EventsClientDaemon.get_instance().get_port()
        self.assertTrue(sig in complist, 'Failed registering client.')
        self.assertTrue(port in complist[sig],
                        'Failed registering client port.')

    def test_client_receives_signal(self):
        """
        Ensure clients can receive signals.
        """
        sig = 7
        flag = Mock()

        events.register(sig, lambda req: flag(req.event))
        request = SignalRequest()
        request.event = sig
        request.content = ""
        request.mac_method = mac_auth.MacMethod.MAC_NONE
        request.mac = ""
        service = RpcService(EventsServerService_Stub, port, 'localhost')
        response = service.signal(request, timeout=1000)
        self.assertTrue(response is not None, 'Did not receive response.')
        time.sleep(0.5)
        flag.assert_called_once_with(sig)

    def test_client_send_signal(self):
        """
        Ensure clients can send signals.
        """
        sig = 8
        response = events.signal(sig)
        self.assertTrue(response.status == response.OK,
                        'Received wrong response status when signaling.')

    def test_client_unregister_all(self):
        """
        Test that the client can unregister all events for one signal.
        """
        sig = CLIENT_UID
        complist = server.registered_clients
        events.register(sig, lambda x: True)
        events.register(sig, lambda x: True)
        time.sleep(0.1)
        events.unregister(sig)
        time.sleep(0.1)
        port = client.EventsClientDaemon.get_instance().get_port()
        self.assertFalse(bool(complist[sig]))
        self.assertTrue(port not in complist[sig])

    def test_client_unregister_by_uid(self):
        """
        Test that the client can unregister an event by uid.
        """
        sig = CLIENT_UID
        complist = server.registered_clients
        events.register(sig, lambda x: True, uid='cbkuid')
        events.register(sig, lambda x: True, uid='cbkuid2')
        time.sleep(0.1)
        events.unregister(sig, uid='cbkuid')
        time.sleep(0.1)
        port = client.EventsClientDaemon.get_instance().get_port()
        self.assertTrue(sig in complist)
        self.assertTrue(len(complist[sig]) == 1)
        self.assertTrue(
            client.registered_callbacks[sig].pop()[0] == 'cbkuid2')
        self.assertTrue(port in complist[sig])

    def test_server_replies_ping(self):
        """
        Ensure server replies to a ping.
        """
        request = PingRequest()
        service = RpcService(EventsServerService_Stub, port, 'localhost')
        response = service.ping(request, timeout=1000)
        self.assertIsNotNone(response)
        self.assertEqual(EventResponse.OK, response.status,
                         'Wrong response status.')

    def test_client_replies_ping(self):
        """
        Ensure clients reply to a ping.
        """
        daemon = client.ensure_client_daemon()
        port = daemon.get_port()
        request = PingRequest()
        service = RpcService(EventsClientService_Stub, port, 'localhost')
        response = service.ping(request, timeout=1000)
        self.assertEqual(EventResponse.OK, response.status,
                         'Wrong response status.')

    def test_server_ping(self):
        """
        Ensure the function from server module pings correctly.
        """
        response = server.ping()
        self.assertIsNotNone(response)
        self.assertEqual(EventResponse.OK, response.status,
                         'Wrong response status.')

    def test_client_ping(self):
        """
        Ensure the function from client module pings correctly.
        """
        daemon = client.ensure_client_daemon()
        response = client.ping(daemon.get_port())
        self.assertIsNotNone(response)
        self.assertEqual(EventResponse.OK, response.status,
                         'Wrong response status.')

    def test_module_ping_server(self):
        """
        Ensure the function from main module pings server correctly.
        """
        response = events.ping_server()
        self.assertIsNotNone(response)
        self.assertEqual(EventResponse.OK, response.status,
                         'Wrong response status.')

    def test_module_ping_client(self):
        """
        Ensure the function from main module pings clients correctly.
        """
        daemon = client.ensure_client_daemon()
        response = events.ping_client(daemon.get_port())
        self.assertIsNotNone(response)
        self.assertEqual(EventResponse.OK, response.status,
                         'Wrong response status.')

    def test_ensure_server_raises_if_port_taken(self):
        """
        Verify that server raises an exception if port is already taken.
        """
        # get a random free port
        while True:
            port = random.randint(1024, 65535)
            try:
                s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                s.connect(('localhost', port))
                s.close()
            except:
                break

        class PortBlocker(threading.Thread):

            def run(self):
                conns = 0
                s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                s.bind(('localhost', port))
                s.setblocking(1)
                s.listen(1)
                while conns < 2:  # blocks until rece
                    conns += 1
                    s.accept()
                s.close()

        # block the port
        taker = PortBlocker()
        taker.start()
        time.sleep(1)  # wait for thread to start.
        self.assertRaises(
            server.PortAlreadyTaken, server.ensure_server, port)

    def test_async_register(self):
        """
        Test asynchronous registering of callbacks.
        """
        flag = Mock()

        # executed after async register, when response is received from server
        def reqcbk(request, response):
            flag(request.event, response.status)

        # callback registered by application
        def callback(request):
            pass

        # passing a callback as reqcbk param makes the call asynchronous
        result = events.register(CLIENT_UID, callback, reqcbk=reqcbk)
        self.assertIsNone(result)
        events.signal(CLIENT_UID)
        time.sleep(1)  # wait for signal to arrive from server
        flag.assert_called_once_with(CLIENT_UID, EventResponse.OK)

    def test_async_signal(self):
        """
        Test asynchronous registering of callbacks.
        """
        flag = Mock()

        # executed after async signal, when response is received from server
        def reqcbk(request, response):
            flag(request.event, response.status)

        # passing a callback as reqcbk param makes the call asynchronous
        result = events.signal(CLIENT_UID, reqcbk=reqcbk)
        self.assertIsNone(result)
        time.sleep(1)  # wait for signal to arrive from server
        flag.assert_called_once_with(CLIENT_UID, EventResponse.OK)

    def test_async_unregister(self):
        """
        Test asynchronous unregistering of callbacks.
        """
        flag = Mock()

        # executed after async signal, when response is received from server
        def reqcbk(request, response):
            flag(request.event, response.status)

        # callback registered by application
        def callback(request):
            pass

        # passing a callback as reqcbk param makes the call asynchronous
        events.register(CLIENT_UID, callback)
        result = events.unregister(CLIENT_UID, reqcbk=reqcbk)
        self.assertIsNone(result)
        time.sleep(1)  # wait for signal to arrive from server
        flag.assert_called_once_with(CLIENT_UID, EventResponse.OK)

    def test_async_ping_server(self):
        """
        Test asynchronous pinging of server.
        """
        flag = Mock()

        # executed after async signal, when response is received from server
        def reqcbk(request, response):
            flag(response.status)

        result = events.ping_server(reqcbk=reqcbk)
        self.assertIsNone(result)
        time.sleep(1)  # wait for response to arrive from server.
        flag.assert_called_once_with(EventResponse.OK)

    def test_async_ping_client(self):
        """
        Test asynchronous pinging of client.
        """
        flag = Mock()

        # executed after async signal, when response is received from server
        def reqcbk(request, response):
            flag(response.status)

        daemon = client.ensure_client_daemon()
        result = events.ping_client(daemon.get_port(), reqcbk=reqcbk)
        self.assertIsNone(result)
        time.sleep(1)  # wait for response to arrive from server.
        flag.assert_called_once_with(EventResponse.OK)



# -*- coding: utf-8 -*-
# server.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
A server for the events mechanism.

A server can receive different kinds of requests from clients:

  1. Registration request: store client port number to be notified when
     a specific signal arrives.

  2. Signal request: redistribute the signal to registered clients.
"""
import logging
import socket


from protobuf.socketrpc import RpcService
from leap.common.events import (
    events_pb2 as proto,
    daemon,
)


logger = logging.getLogger(__name__)


SERVER_PORT = 8090

# the `registered_clients` dictionary below should have the following
# format:
#
#     { event_signal: [ port, ... ], ... }
#
registered_clients = {}


class PortAlreadyTaken(Exception):
    """
    Raised when trying to open a server in a port that is already taken.
    """
    pass


def ensure_server(port=SERVER_PORT):
    """
    Make sure the server is running on the given port.

    Attempt to connect to given local port. Upon success, assume that the
    events server has already been started. Upon failure, start events server.

    :param port: the port in which server should be listening
    :type port: int

    :return: the daemon instance or nothing
    :rtype: EventsServerDaemon or None

    :raise PortAlreadyTaken: Raised if C{port} is already taken by something
                             that is not an events server.
    """
    try:
        # check if port is available
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(('localhost', port))
        s.close()
        # port is taken, check if there's a server running there
        response = ping(port=port, timeout=1000)
        if response is not None and response.status == proto.EventResponse.OK:
            logger.info('A server is already running on port %d.', port)
            return
        # port is taken, and not by an events server
        logger.warning(
            'Port %d is taken by something not an events server.', port)
        raise PortAlreadyTaken(port)
    except socket.error:
        # port is available, run a server
        logger.info('Launching server on port %d.', port)
        return EventsServerDaemon.ensure(port)


def ping(port=SERVER_PORT, reqcbk=None, timeout=1000):
    """
    Ping the server.

    :param port: the port in which server should be listening
    :type port: int
    :param reqcbk: a callback to be called when a response from server is
                   received
    :type reqcbk: function(proto.PingRequest, proto.EventResponse)
    :param timeout: the timeout for synch calls
    :type timeout: int

    :return: the response from server for synch calls or nothing for asynch
             calls.
    :rtype: leap.common.events.events_pb2.EventsResponse or None
    """
    request = proto.PingRequest()
    service = RpcService(
        proto.EventsServerService_Stub,
        port,
        'localhost')
    logger.debug("Pinging server in port %d..." % port)
    return service.ping(request, callback=reqcbk, timeout=timeout)


class EventsServerService(proto.EventsServerService):
    """
    Service for receiving events in clients.
    """

    def register(self, controller, request, done):
        """
        Register a client port to be signaled when specific events come in.

        :param controller: used to mediate a single method call
        :type controller: protobuf.socketrpc.controller.SocketRpcController
        :param request: the request received from the client
        :type request: leap.common.events.events_pb2.RegisterRequest
        :param done: callback to be called when done
        :type done: protobuf.socketrpc.server.Callback
        """
        logger.info("Received registration request: %s..." % str(request)[:40])
        # add client port to signal list
        if request.event not in registered_clients:
            registered_clients[request.event] = set([])
        registered_clients[request.event].add(request.port)
        # send response back to client

        logger.debug('sending response back')
        response = proto.EventResponse()
        response.status = proto.EventResponse.OK
        done.run(response)

    def unregister(self, controller, request, done):
        """
        Unregister a client port so it will not be signaled when specific
        events come in.

        :param controller: used to mediate a single method call
        :type controller: protobuf.socketrpc.controller.SocketRpcController
        :param request: the request received from the client
        :type request: leap.common.events.events_pb2.RegisterRequest
        :param done: callback to be called when done
        :type done: protobuf.socketrpc.server.Callback
        """
        logger.info(
            "Received unregistration request: %s..." % str(request)[:40])
        # remove client port from signal list
        response = proto.EventResponse()
        if request.event in registered_clients:
            try:
                registered_clients[request.event].remove(request.port)
                response.status = proto.EventResponse.OK
            except KeyError:
                response.status = proto.EventsResponse.ERROR
                response.result = 'Port %d not registered.' % request.port
        # send response back to client
        logger.debug('sending response back')
        done.run(response)

    def signal(self, controller, request, done):
        """
        Perform an RPC call to signal all clients registered to receive a
        specific signal.

        :param controller: used to mediate a single method call
        :type controller: protobuf.socketrpc.controller.SocketRpcController
        :param request: the request received from the client
        :type request: leap.common.events.events_pb2.SignalRequest
        :param done: callback to be called when done
        :type done: protobuf.socketrpc.server.Callback
        """
        logger.debug('Received signal from client: %s...', str(request)[:40])
        # send signal to all registered clients
        # TODO: verify signal auth
        if request.event in registered_clients:
            for port in registered_clients[request.event]:

                def callback(req, resp):
                    logger.debug("Signal received by " + str(port))

                service = RpcService(proto.EventsClientService_Stub,
                                     port, 'localhost')
                service.signal(request, callback=callback)
        # send response back to client
        response = proto.EventResponse()
        response.status = proto.EventResponse.OK
        done.run(response)

    def ping(self, controller, request, done):
        """
        Reply to a ping request.

        :param controller: used to mediate a single method call
        :type controller: protobuf.socketrpc.controller.SocketRpcController
        :param request: the request received from the client
        :type request: leap.common.events.events_pb2.RegisterRequest
        :param done: callback to be called when done
        :type done: protobuf.socketrpc.server.Callback
        """
        logger.debug("Received ping request, sending response.")
        response = proto.EventResponse()
        response.status = proto.EventResponse.OK
        done.run(response)


class EventsServerDaemon(daemon.EventsSingletonDaemon):
    """
    Singleton class for starting an events server daemon.
    """

    @classmethod
    def ensure(cls, port):
        """
        Make sure the daemon is running on the given port.

        :param port: the port in which the daemon should listen
        :type port: int

        :return: a daemon instance
        :rtype: EventsServerDaemon
        """
        return cls.ensure_service(port, EventsServerService())



# -*- coding: utf-8 -*-
# leap.common.testing.https_server.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
A simple HTTPS server to be used in tests
"""
from BaseHTTPServer import HTTPServer
import os
import ssl
import SocketServer
import threading
import unittest

_where = os.path.split(__file__)[0]


def where(filename):
    return os.path.join(_where, filename)


class HTTPSServer(HTTPServer):
    def server_bind(self):
        SocketServer.TCPServer.server_bind(self)
        self.socket = ssl.wrap_socket(
            self.socket, server_side=True,
            certfile=where("leaptestscert.pem"),
            keyfile=where("leaptestskey.pem"),
            ca_certs=where("cacert.pem"),
            ssl_version=ssl.PROTOCOL_SSLv23)


class TestServerThread(threading.Thread):
    def __init__(self, test_object, request_handler):
        threading.Thread.__init__(self)
        self.request_handler = request_handler
        self.test_object = test_object

    def run(self):
        self.server = HTTPSServer(('localhost', 0), self.request_handler)
        host, port = self.server.socket.getsockname()
        self.test_object.HOST, self.test_object.PORT = host, port
        self.test_object.server_started.set()
        self.test_object = None
        try:
            self.server.serve_forever(0.05)
        finally:
            self.server.server_close()

    def stop(self):
        self.server.shutdown()


class BaseHTTPSServerTestCase(unittest.TestCase):
    """
    derived classes need to implement a request_handler
    """
    def setUp(self):
        self.server_started = threading.Event()
        self.thread = TestServerThread(self, self.request_handler)
        self.thread.start()
        self.server_started.wait()

    def tearDown(self):
        self.thread.stop()

    def get_server(self):
        host, port = self.HOST, self.PORT
        if host == "127.0.0.1":
            host = "localhost"
        return "%s:%s" % (host, port)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# decorators.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Useful decorators.
"""
import collections
import datetime
import functools
import logging

try:
    from twisted.internet import defer
except ImportError:
    class defer:
        class Deferred:
            pass

logger = logging.getLogger(__name__)


class _memoized(object):
    """
    Decorator.

    Caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned
    (not reevaluated).
    """

    # cache invalidation time, in seconds
    CACHE_INVALIDATION_DELTA = 1800

    def __init__(self, func, ignore_kwargs=None, is_method=False,
                 invalidation=None):

        """
        :param ignore_kwargs: If True, ignore all kwargs.
                              If tuple, ignore those kwargs.
        :type ignore_kwargs: bool, tuple or None
        :param is_method: whether the decorated function is a method.
                          (ignores the self argument if so).
        :type is_method: True
        """
        self.ignore_kwargs = ignore_kwargs if ignore_kwargs else []
        self.is_method = is_method
        self.func = func

        if invalidation:
            self.CACHE_INVALIDATION_DELTA = invalidation

        # TODO should put bounds to the cache dict so we do not
        # consume a huge amount of memory.
        self.cache = {}
        self.cache_ts = {}
        self.is_deferred = {}

    def __call__(self, *args, **kwargs):
        """
        Executes the call.

        :tyoe args: tuple
        :type kwargs: dict
        """
        key = self._build_key(*args, **kwargs)
        if not isinstance(key, collections.Hashable):
            # uncacheable. a list, for instance.
            # better to not cache than blow up.
            logger.warning("Key is not hashable, bailing out!")
            return self.func(*args, **kwargs)

        if key in self.cache:
            if self._is_cache_still_valid(key):
                logger.debug("Got value from cache...")
                value = self._get_value(key)
                return self._ret_or_raise(value)
            else:
                logger.debug("Cache is invalid, evaluating again...")

        # no cache, or cache invalid
        try:
            value = self.func(*args, **kwargs)
        except Exception as exc:
            logger.error("Exception while calling function: %r" % (exc,))
            value = exc

        if isinstance(value, defer.Deferred):
            value.addBoth(self._store_deferred_value(key))
        else:
            self.cache[key] = value
            self.is_deferred[key] = False
        self.cache_ts[key] = datetime.datetime.now()
        return self._ret_or_raise(value)

    def _build_key(self, *args, **kwargs):
        """
        Build key from the function arguments
        """
        if self.is_method:
            # forget about `self` as key
            key_args = args[1:]
        else:
            key_args = args

        if self.ignore_kwargs is True:
            key = key_args
        else:
            key = (key_args, frozenset(
                [(k, v) for k, v in kwargs.items()
                 if k not in self.ignore_kwargs]))
        return key

    def _get_value(self, key):
        """
        Get a value form cache for a key
        """
        if self.is_deferred[key]:
            value = self.cache[key]
            # it produces an errback if value is Failure
            return defer.succeed(value)
        else:
            return self.cache[key]

    def _ret_or_raise(self, value):
        """
        Returns the value except if it is an exception,
        in which case it's raised.
        """
        if isinstance(value, Exception):
            raise value
        return value

    def _is_cache_still_valid(self, key, now=datetime.datetime.now):
        """
        Returns True if the cache value is still valid, False otherwise.

        For now, this happen if less than CACHE_INVALIDATION_DELTA seconds
        have passed from the time in which we recorded the cached value.

        :param key: the key to lookup in the cache
        :type key: hashable
        :param now: a callable that returns a datetime object. override
                    for dependency injection during testing.
        :type now: callable
        :rtype: bool
        """
        cached_ts = self.cache_ts[key]
        delta = datetime.timedelta(seconds=self.CACHE_INVALIDATION_DELTA)
        return (now() - cached_ts) < delta

    def _store_deferred_value(self, key):
        """
        Returns a callback to store values from deferreds
        """
        def callback(value):
            self.cache[key] = value
            self.is_deferred[key] = True
            return value
        return callback

    def __repr__(self):
        """
        Return the function's docstring.
        """
        return self.func.__doc__

    def __get__(self, obj, objtype):
        """
        Support instance methods.
        """
        return functools.partial(self.__call__, obj)


def memoized_method(function=None, ignore_kwargs=None,
                    invalidation=_memoized.CACHE_INVALIDATION_DELTA):
    """
    Wrap _memoized to allow for deferred calling

    :type function: callable, or None.
    :type ignore_kwargs: None, True or tuple.
    :type invalidation: int seconds.
    """
    if function:
        return _memoized(function, is_method=True, invalidation=invalidation)
    else:
        def wrapper(function):
            return _memoized(
                function, ignore_kwargs=ignore_kwargs, is_method=True,
                invalidation=invalidation)
        return wrapper



# -*- coding: utf-8 -*-
# files.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Implements file helper methods
"""

import errno
import logging
import os
import stat
import time

logger = logging.getLogger(__name__)


def check_and_fix_urw_only(cert):
    """
    Test for 600 mode and try to set it if anything different found

    Might raise OSError

    :param cert: Certificate path
    :type cert: str
    """
    mode = stat.S_IMODE(os.stat(cert).st_mode)

    if mode != int('600', 8):
        try:
            logger.warning('Bad permission on %s attempting to set 600' %
                           (cert,))
            os.chmod(cert, stat.S_IRUSR | stat.S_IWUSR)
        except OSError:
            logger.error('Error while trying to chmod 600 %s' %
                         cert)
            raise


def get_mtime(filename):
    """
    Returns the modified time or None if the file doesn't exist

    :param filename: path to check
    :type filename: str

    :rtype: str
    """
    try:
        mtime = time.ctime(os.path.getmtime(filename)) + " GMT"
        return mtime
    except OSError:
        return None


def mkdir_p(path):
    """
    Creates the path and all the intermediate directories that don't
    exist

    Might raise OSError

    :param path: path to create
    :type path: str
    """
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


# Twisted implementation of which
def which(name, flags=os.X_OK, path_extension="/usr/sbin:/sbin"):
    """
    Search PATH for executable files with the given name.

    On newer versions of MS-Windows, the PATHEXT environment variable will be
    set to the list of file extensions for files considered executable. This
    will normally include things like ".EXE". This fuction will also find files
    with the given name ending with any of these extensions.

    On MS-Windows the only flag that has any meaning is os.F_OK. Any other
    flags will be ignored.

    :type name: C{str}
    :param name: The name for which to search.

    :type flags: C{int}
    :param flags: Arguments to L{os.access}.

    :rtype: C{list}
    :param: A list of the full paths to files found, in the
    order in which they were found.
    """

    result = []
    exts = filter(None, os.environ.get('PATHEXT', '').split(os.pathsep))
    path = os.environ.get('PATH', None)
    path = path_extension + os.pathsep + path
    if path is None:
        return []
    parts = path.split(os.pathsep)
    for p in parts:
        p = os.path.join(p, name)
        if os.access(p, flags):
            result.append(p)
        for e in exts:
            pext = p + e
            if os.access(pext, flags):
                result.append(pext)
    return result



# -*- coding: utf-8 -*-
# client.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
The client end point of the events mechanism.

Clients are the communicating parties of the events mechanism. They
communicate by sending messages to a server, which in turn redistributes
messages to other clients.

When a client registers a callback for a given signal, it also tells the
server that it wants to be notified whenever signals of that type are sent by
some other client.
"""

import logging


from protobuf.socketrpc import RpcService

from leap.common.events import events_pb2 as proto
from leap.common.events import server
from leap.common.events import daemon
from leap.common.events import mac_auth


logger = logging.getLogger(__name__)


# the `registered_callbacks` dictionary below should have the following
# format:
#
#     { event_signal: [ (uid, callback), ... ], ... }
#
registered_callbacks = {}


class CallbackAlreadyRegistered(Exception):
    """
    Raised when trying to register an already registered callback.
    """
    pass


def ensure_client_daemon():
    """
    Ensure the client daemon is running and listening for incoming
    messages.

    :return: the daemon instance
    :rtype: EventsClientDaemon
    """
    import time
    daemon = EventsClientDaemon.ensure(0)
    logger.debug('ensure client daemon')

    # Because we use a random port we want to wait until a port is assigned to
    # local client daemon.

    while not (EventsClientDaemon.get_instance() and
               EventsClientDaemon.get_instance().get_port()):
        time.sleep(0.1)
    return daemon


def register(signal, callback, uid=None, replace=False, reqcbk=None,
             timeout=1000):
    """
    Registers a callback to be called when a specific signal event is
    received.

    Will timeout after timeout ms if response has not been received. The
    timeout arg is only used for asynch requests. If a reqcbk callback has
    been supplied the timeout arg is not used. The response value will be
    returned for a synch request but nothing will be returned for an asynch
    request.

    :param signal: the signal that causes the callback to be launched
    :type signal: int (see the `events.proto` file)
    :param callback: the callback to be called when the signal is received
    :type callback: function(leap.common.events.events_pb2.SignalRequest)
    :param uid: a unique id for the callback
    :type uid: int
    :param replace: should an existent callback with same uid be replaced?
    :type replace: bool
    :param reqcbk: a callback to be called when a response from server is
                   received
    :type reqcbk: function(proto.RegisterRequest, proto.EventResponse)
    :param timeout: the timeout for synch calls
    :type timeout: int

    Might raise a CallbackAlreadyRegistered exception if there's already a
    callback identified by the given uid and replace is False.

    :return: the response from server for synch calls or nothing for asynch
             calls.
    :rtype: leap.common.events.events_pb2.EventsResponse or None
    """
    ensure_client_daemon()  # so we can receive registered signals
    # register callback locally
    if signal not in registered_callbacks:
        registered_callbacks[signal] = []
    cbklist = registered_callbacks[signal]

    # TODO should check that the callback has the right
    # number of arguments.

    if uid and filter(lambda (x, y): x == uid, cbklist):
        if not replace:
            raise CallbackAlreadyRegistered()
        else:
            registered_callbacks[signal] = filter(lambda(x, y): x != uid,
                                                  cbklist)
    registered_callbacks[signal].append((uid, callback))
    # register callback on server
    request = proto.RegisterRequest()
    request.event = signal
    request.port = EventsClientDaemon.get_instance().get_port()
    request.mac_method = mac_auth.MacMethod.MAC_NONE
    request.mac = ""
    service = RpcService(proto.EventsServerService_Stub,
                         server.SERVER_PORT, 'localhost')
    logger.debug(
        "Sending registration request to server on port %s: %s",
        server.SERVER_PORT,
        str(request)[:40])
    return service.register(request, callback=reqcbk, timeout=timeout)


def unregister(signal, uid=None, reqcbk=None, timeout=1000):
    """
    Unregister a callback.

    If C{uid} is specified, unregisters only the callback identified by that
    unique id. Otherwise, unregisters all callbacks

    :param signal: the signal that causes the callback to be launched
    :type signal: int (see the `events.proto` file)
    :param uid: a unique id for the callback
    :type uid: int
    :param reqcbk: a callback to be called when a response from server is
                   received
    :type reqcbk: function(proto.UnregisterRequest, proto.EventResponse)
    :param timeout: the timeout for synch calls
    :type timeout: int

    :return: the response from server for synch calls or nothing for asynch
             calls or None if no callback is registered for that signal or
             uid.
    :rtype: leap.common.events.events_pb2.EventsResponse or None
    """
    if signal not in registered_callbacks or not registered_callbacks[signal]:
        logger.warning("No callback registered for signal %d." % signal)
        return None
    # unregister callback locally
    cbklist = registered_callbacks[signal]
    if uid is not None:
        if filter(lambda (cbkuid, _): cbkuid == uid, cbklist) == []:
            logger.warning("No callback registered for uid %d." % st)
            return None
        registered_callbacks[signal] = filter(lambda(x, y): x != uid, cbklist)
    else:
        # exclude all callbacks for given signal
        registered_callbacks[signal] = []
    # unregister port in server if there are no more callbacks for this signal
    if not registered_callbacks[signal]:
        request = proto.UnregisterRequest()
        request.event = signal
        request.port = EventsClientDaemon.get_instance().get_port()
        request.mac_method = mac_auth.MacMethod.MAC_NONE
        request.mac = ""
        service = RpcService(proto.EventsServerService_Stub,
                             server.SERVER_PORT, 'localhost')
        logger.info(
            "Sending unregistration request to server on port %s: %s",
            server.SERVER_PORT,
            str(request)[:40])
        return service.unregister(request, callback=reqcbk, timeout=timeout)


def signal(signal, content="", mac_method="", mac="", reqcbk=None,
           timeout=1000):
    """
    Send `signal` event to events server.

    Will timeout after timeout ms if response has not been received. The
    timeout arg is only used for asynch requests.  If a reqcbk callback has
    been supplied the timeout arg is not used. The response value will be
    returned for a synch request but nothing will be returned for an asynch
    request.

    :param signal: the signal that causes the callback to be launched
    :type signal: int (see the `events.proto` file)
    :param content: the contents of the event signal
    :type content: str
    :param mac_method: the method used for auth mac
    :type mac_method: str
    :param mac: the content of the auth mac
    :type mac: str
    :param reqcbk: a callback to be called when a response from server is
                   received
    :type reqcbk: function(proto.SignalRequest, proto.EventResponse)
    :param timeout: the timeout for synch calls
    :type timeout: int

    :return: the response from server for synch calls or nothing for asynch
             calls.
    :rtype: leap.common.events.events_pb2.EventsResponse or None
    """
    request = proto.SignalRequest()
    request.event = signal
    request.content = content
    request.mac_method = mac_method
    request.mac = mac
    service = RpcService(proto.EventsServerService_Stub, server.SERVER_PORT,
                         'localhost')
    logger.debug("Sending signal to server: %s", str(request)[:40])
    return service.signal(request, callback=reqcbk, timeout=timeout)


def ping(port, reqcbk=None, timeout=1000):
    """
    Ping a client running in C{port}.

    :param port: the port in which the client should be listening
    :type port: int
    :param reqcbk: a callback to be called when a response from client is
                   received
    :type reqcbk: function(proto.PingRequest, proto.EventResponse)
    :param timeout: the timeout for synch calls
    :type timeout: int

    :return: the response from client for synch calls or nothing for asynch
             calls.
    :rtype: leap.common.events.events_pb2.EventsResponse or None
    """
    request = proto.PingRequest()
    service = RpcService(
        proto.EventsClientService_Stub,
        port,
        'localhost')
    logger.debug("Pinging a client in port %d..." % port)
    return service.ping(request, callback=reqcbk, timeout=timeout)


class EventsClientService(proto.EventsClientService):
    """
    Service for receiving signal events in clients.
    """

    def __init__(self):
        proto.EventsClientService.__init__(self)

    def signal(self, controller, request, done):
        """
        Receive a signal and run callbacks registered for that signal.

        This method is called whenever a signal request is received from
        server.

        :param controller: used to mediate a single method call
        :type controller: protobuf.socketrpc.controller.SocketRpcController
        :param request: the request received from the client
        :type request: leap.common.events.events_pb2.SignalRequest
        :param done: callback to be called when done
        :type done: protobuf.socketrpc.server.Callback
        """
        logger.debug('Received signal from server: %s...' % str(request)[:40])

        # run registered callbacks
        # TODO: verify authentication using mac in incoming message
        if request.event in registered_callbacks:
            for (_, cbk) in registered_callbacks[request.event]:
                # callbacks should be prepared to receive a
                # events_pb2.SignalRequest.
                cbk(request)

        # send response back to server
        response = proto.EventResponse()
        response.status = proto.EventResponse.OK
        done.run(response)

    def ping(self, controller, request, done):
        """
        Reply to a ping request.

        :param controller: used to mediate a single method call
        :type controller: protobuf.socketrpc.controller.SocketRpcController
        :param request: the request received from the client
        :type request: leap.common.events.events_pb2.RegisterRequest
        :param done: callback to be called when done
        :type done: protobuf.socketrpc.server.Callback
        """
        logger.debug("Received ping request, sending response.")
        response = proto.EventResponse()
        response.status = proto.EventResponse.OK
        done.run(response)


class EventsClientDaemon(daemon.EventsSingletonDaemon):
    """
    A daemon that listens for incoming events from server.
    """

    @classmethod
    def ensure(cls, port):
        """
        Make sure the daemon is running on the given port.

        :param port: the port in which the daemon should listen
        :type port: int

        :return: a daemon instance
        :rtype: EventsClientDaemon
        """
        return cls.ensure_service(port, EventsClientService())



# -*- coding: utf-8 -*-
# check.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Set of functions to help checking situations
"""

import logging
import traceback


logger = logging.getLogger(__name__)


def leap_assert(condition, message=""):
    """
    Asserts the condition and displays the message if that's not
    met. It also logs the error and its backtrace.

    :param condition: condition to check
    :type condition: bool
    :param message: message to display if the condition isn't met
    :type message: str
    """
    if not condition:
        logger.error("Bug: %s" % (message,))
        try:
            for formatted_line in traceback.format_list(
                    traceback.extract_stack()[:-1]):
                for line in formatted_line.split("\n"):
                    if len(line.strip()) > 0:
                        logger.error(line)
        except Exception as e:
            logger.error("Bug in leap_assert: %r" % (e,))
    assert condition, message


def leap_assert_type(var, expectedType):
    """
    Helper assert check for a variable's expected type

    :param var: variable to check
    :type var: any
    :param expectedType: type to check agains
    :type expectedType: type
    """
    leap_assert(isinstance(var, expectedType),
                "Expected type %r instead of %r" %
                (expectedType, type(var)))


def leap_check(condition, message="", exception=Exception):
    """
    Asserts the condition and displays the message if that's not
    met. It also logs the error and its backtrace.

    :param condition: condition to check
    :type condition: bool
    :param message: message to display if the condition isn't met
    :type message: str
    :param exception: the exception to raise in case of condition being false.
    :type exception: Exception
    """
    if not condition:
        logger.error(message)
        raise exception(message)



# -*- coding: utf-8 -*-
# daemon.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
A singleton daemon for running RPC services using protobuf.socketrpc.
"""


import logging
import threading


from protobuf.socketrpc.server import (
    SocketRpcServer,
    ThreadedTCPServer,
    SocketHandler,
)


logger = logging.getLogger(__name__)


class ServiceAlreadyRunningException(Exception):
    """
    Raised whenever a service is already running in this process but someone
    attemped to start it in a different port.
    """


class EventsRpcServer(SocketRpcServer):
    """
    RPC server used in server and client interfaces to receive messages.
    """

    def __init__(self, port, host='localhost'):
        """
        Initialize a RPC server.

        :param port: the port in which to listen for incoming messages
        :type port: int
        :param host: the address to bind to
        :type host: str
        """
        SocketRpcServer.__init__(self, port, host)
        self._server = None

    def run(self):
        """
        Run the server.
        """
        logger.info('Running server on port %d.' % self.port)
        # parent implementation does not hold the server instance, so we do it
        # here.
        self._server = ThreadedTCPServer((self.host, self.port),
                                         SocketHandler, self)
        # if we chose to use a random port, fetch the port number info.
        if self.port is 0:
            self.port = self._server.socket.getsockname()[1]
        self._server.serve_forever()

    def stop(self):
        """
        Stop the server.
        """
        self._server.shutdown()


class EventsSingletonDaemon(threading.Thread):
    """
    Singleton class for for launching and terminating a daemon.

    This class is used so every part of the mechanism that needs to listen for
    messages can launch its own daemon (thread) to do the job.
    """

    # Singleton instance
    __instance = None

    def __new__(cls, *args, **kwargs):
        """
        Return a singleton instance if it exists or create and initialize one.
        """
        if len(args) is not 2:
            raise TypeError("__init__() takes exactly 2 arguments (%d given)"
                            % len(args))
        if cls.__instance is None:
            cls.__instance = object.__new__(
                EventsSingletonDaemon)
            cls.__initialize(cls.__instance, args[0], args[1])
        return cls.__instance

    @staticmethod
    def __initialize(self, port, service):
        """
        Initialize a singleton daemon.

        This is a static method disguised as instance method that actually
        does the initialization of the daemon instance.

        :param port: the port in which to listen for incoming messages
        :type port: int
        :param service: the service to provide in this daemon
        :type service: google.protobuf.service.Service
        """
        threading.Thread.__init__(self)
        self._port = port
        self._service = service
        self._server = EventsRpcServer(self._port)
        self._server.registerService(self._service)
        self.daemon = True

    def __init__(self):
        """
        Singleton placeholder initialization method.

        Initialization is made in __new__ so we can always return the same
        instance upon object creation.
        """
        pass

    @classmethod
    def ensure(cls, port):
        """
        Make sure the daemon instance is running.

        Each implementation of this method should call `self.ensure_service`
        with the appropriate service from the `events.proto` definitions, and
        return the daemon instance.

        :param port: the port in which the daemon should be listening
        :type port: int

        :return: a daemon instance
        :rtype: EventsSingletonDaemon
        """
        raise NotImplementedError(self.ensure)

    @classmethod
    def ensure_service(cls, port, service):
        """
        Start the singleton instance if not already running.

        Might return ServiceAlreadyRunningException

        :param port: the port in which the daemon should be listening
        :type port: int

        :return: a daemon instance
        :rtype: EventsSingletonDaemon
        """
        daemon = cls(port, service)
        if not daemon.is_alive():
            daemon.start()
        elif port and port != cls.__instance._port:
            # service is running in this process but someone is trying to
            # start it in another port
            raise ServiceAlreadyRunningException(
                "Service is already running in this process on port %d."
                % self.__instance._port)
        return daemon

    @classmethod
    def get_instance(cls):
        """
        Retrieve singleton instance of this daemon.

        :return: a daemon instance
        :rtype: EventsSingletonDaemon
        """
        return cls.__instance

    def run(self):
        """
        Run the server.
        """
        self._server.run()

    def stop(self):
        """
        Stop the daemon.
        """
        self._server.stop()

    def get_port(self):
        """
        Retrieve the value of the port to which the service running in this
        daemon is binded to.

        :return: the port to which the daemon is binded to
        :rtype: int
        """
        if self._port is 0:
            self._port = self._server.port
        return self._port



# -*- coding: utf-8 -*-
# certs.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Implements cert checks and helpers
"""

import os
import time
import logging

from OpenSSL import crypto
from dateutil.parser import parse as dateparse

from leap.common.check import leap_assert

logger = logging.getLogger(__name__)


def get_cert_from_string(string):
    """
    Returns the x509 from the contents of this string

    :param string: certificate contents as downloaded
    :type string: str

    :return: x509 or None
    """
    leap_assert(string, "We need something to load")

    x509 = None
    try:
        x509 = crypto.load_certificate(crypto.FILETYPE_PEM, string)
    except Exception as e:
        logger.error("Something went wrong while loading the certificate: %r"
                     % (e,))
    return x509


def get_privatekey_from_string(string):
    """
    Returns the private key from the contents of this string

    :param string: private key contents as downloaded
    :type string: str

    :return: private key or None
    """
    leap_assert(string, "We need something to load")

    pkey = None
    try:
        pkey = crypto.load_privatekey(crypto.FILETYPE_PEM, string)
    except Exception as e:
        logger.error("Something went wrong while loading the certificate: %r"
                     % (e,))
    return pkey


def get_digest(cert_data, method):
    """
    Returns the digest for the cert_data using the method specified

    :param cert_data: certificate data in string form
    :type cert_data: str
    :param method: method to be used for digest
    :type method: str

    :rtype: str
    """
    x509 = get_cert_from_string(cert_data)
    digest = x509.digest(method).replace(":", "").lower()

    return digest


def can_load_cert_and_pkey(string):
    """
    Loads certificate and private key from a buffer, returns True if
    everything went well, False otherwise

    :param string: buffer containing the cert and private key
    :type string: str or any kind of buffer

    :rtype: bool
    """
    can_load = True

    try:
        cert = get_cert_from_string(string)
        key = get_privatekey_from_string(string)

        leap_assert(cert, 'The certificate could not be loaded')
        leap_assert(key, 'The private key could not be loaded')
    except Exception as e:
        can_load = False
        logger.error("Something went wrong while trying to load "
                     "the certificate: %r" % (e,))

    return can_load


def is_valid_pemfile(cert):
    """
    Checks that the passed string is a valid pem certificate

    :param cert: String containing pem content
    :type cert: str

    :rtype: bool
    """
    leap_assert(cert, "We need a cert to load")

    return can_load_cert_and_pkey(cert)


def get_cert_time_boundaries(certfile):
    """
    Returns the time boundaries for the certificate saved in certfile

    :param certfile: path to certificate
    :type certfile: str

    :rtype: tuple (from, to)
    """
    cert = get_cert_from_string(certfile)
    leap_assert(cert, 'There was a problem loading the certificate')

    fromts, tots = (cert.get_notBefore(), cert.get_notAfter())
    from_, to_ = map(
        lambda ts: time.gmtime(time.mktime(dateparse(ts).timetuple())),
        (fromts, tots))
    return from_, to_


def should_redownload(certfile, now=time.gmtime):
    """
    Returns True if any of the checks don't pass, False otherwise

    :param certfile: path to certificate
    :type certfile: str
    :param now: current date function, ONLY USED FOR TESTING

    :rtype: bool
    """
    exists = os.path.isfile(certfile)

    if not exists:
        return True

    certdata = None
    try:
        with open(certfile, "r") as f:
            certdata = f.read()
            if not is_valid_pemfile(certdata):
                return True
    except:
        return True

    valid_from, valid_to = get_cert_time_boundaries(certdata)

    if not (valid_from < now() < valid_to):
        return True

    return False



# -*- coding: utf-8 -*-
# __init__.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
This is an events mechanism that uses a server to allow for signaling of
events between clients.

Application components should use the interface available in this file to
register callbacks to be executed upon receiving specific signals, and to send
signals to other components.

To register a callback to be executed when a specific event is signaled, use
leap.common.events.register():

>>> from leap.common.events import register
>>> from leap.common.events import events_pb2 as proto
>>> register(proto.CLIENT_UID, lambda req: do_something(req))

To signal an event, use leap.common.events.signal():

>>> from leap.common.events import signal
>>> from leap.common.events import events_pb2 as proto
>>> signal(proto.CLIENT_UID)


NOTE ABOUT SYNC/ASYNC REQUESTS:

Clients always communicate with the server, and never between themselves.
When a client registers a callback for an event, the callback is saved locally
in the client and the server stores the client socket port in a list
associated with that event. When a client signals an event, the server
forwards the signal to all registered client ports, and then each client
executes its callbacks associated with that event locally.

Each RPC call from a client to the server is followed by a response from the
server to the client. Calls to register() and signal() (and all other RPC
calls) can be synchronous or asynchronous meaning if they will or not wait for
the server's response before returning.

This mechanism was built on top of protobuf.socketrpc, and because of this RPC
calls are made synchronous or asynchronous in the following way:

  * If RPC calls receive a parameter called `reqcbk`, then the call is made
    asynchronous. That means that:

        - an eventual `timeout` parameter is not used,
        - the call returns immediatelly with value None, and
        - the `reqcbk` callback is executed asynchronously upon the arrival of
          a response from the server.

  * Otherwise, if the `reqcbk` parameter is None, then the call is made in a
    synchronous manner:

        - if a response from server arrives within `timeout` milliseconds, the
          RPC call returns it;
        - otherwise, the call returns None.
"""

import logging
import socket


from leap.common.events import (
    events_pb2 as proto,
    server,
    client,
    daemon,
)


logger = logging.getLogger(__name__)


def register(signal, callback, uid=None, replace=False, reqcbk=None,
             timeout=1000):
    """
    Register a callback to be called when the given signal is received.

    Will timeout after timeout ms if response has not been received. The
    timeout arg is only used for asynch requests. If a reqcbk callback has
    been supplied the timeout arg is not used. The response value will be
    returned for a synch request but nothing will be returned for an asynch
    request.

    :param signal: the signal that causes the callback to be launched
    :type signal: int (see the `events.proto` file)
    :param callback: the callback to be called when the signal is received
    :type callback: function
    :param uid: a unique id for the callback
    :type uid: int
    :param replace: should an existent callback with same uid be replaced?
    :type replace: bool
    :param reqcbk: a callback to be called when a response from server is
                   received
    :type reqcbk: function(leap.common.events.events_pb2.EventResponse)
    :param timeout: the timeout for synch calls
    :type timeout: int

    :return: the response from server for synch calls or nothing for asynch
        calls.
    :rtype: leap.common.events.events_pb2.EventsResponse or None
    """
    return client.register(signal, callback, uid, replace, reqcbk, timeout)


def unregister(signal, uid=None, reqcbk=None, timeout=1000):
    """
    Unregister a callback.

    If C{uid} is specified, unregisters only the callback identified by that
    unique id. Otherwise, unregisters all callbacks registered for C{signal}.

    :param signal: the signal that causes the callback to be launched
    :type signal: int (see the `events.proto` file)
    :param uid: a unique id for the callback
    :type uid: int
    :param reqcbk: a callback to be called when a response from server is
                   received
    :type reqcbk: function(proto.UnregisterRequest, proto.EventResponse)
    :param timeout: the timeout for synch calls
    :type timeout: int

    :return: the response from server for synch calls or nothing for asynch
             calls.
    :rtype: leap.common.events.events_pb2.EventsResponse or None
    """
    return client.unregister(signal, uid, reqcbk, timeout)


def signal(signal, content="", mac_method="", mac="", reqcbk=None,
           timeout=1000):
    """
    Send `signal` event to events server.

    Will timeout after timeout ms if response has not been received. The
    timeout arg is only used for asynch requests.  If a reqcbk callback has
    been supplied the timeout arg is not used. The response value will be
    returned for a synch request but nothing will be returned for an asynch
    request.

    :param signal: the signal that causes the callback to be launched
    :type signal: int (see the `events.proto` file)
    :param content: the contents of the event signal
    :type content: str
    :param mac_method: the method used to auth mac
    :type mac_method: str
    :param mac: the content of the auth mac
    :type mac: str
    :param reqcbk: a callback to be called when a response from server is
                   received
    :type reqcbk: function(proto.SignalRequest, proto.EventResponse)
    :param timeout: the timeout for synch calls
    :type timeout: int

    :return: the response from server for synch calls or nothing for asynch
             calls.
    :rtype: leap.common.events.events_pb2.EventsResponse or None
    """
    return client.signal(signal, content, mac_method, mac, reqcbk, timeout)

def ping_client(port, reqcbk=None, timeout=1000):
    """
    Ping a client running in C{port}.

    :param port: the port in which the client should be listening
    :type port: int
    :param reqcbk: a callback to be called when a response from client is
                   received
    :type reqcbk: function(proto.PingRequest, proto.EventResponse)
    :param timeout: the timeout for synch calls
    :type timeout: int
    """
    return client.ping(port, reqcbk=reqcbk, timeout=timeout)


def ping_server(port=server.SERVER_PORT, reqcbk=None, timeout=1000):
    """
    Ping the server.

    :param port: the port in which server should be listening
    :type port: int
    :param reqcbk: a callback to be called when a response from server is
                   received
    :type reqcbk: function(proto.PingRequest, proto.EventResponse)
    :param timeout: the timeout for synch calls
    :type timeout: int
    """
    return server.ping(port, reqcbk=reqcbk, timeout=timeout)



# Generated by the protocol buffer compiler.  DO NOT EDIT!

from google.protobuf import descriptor
from google.protobuf import message
from google.protobuf import reflection
from google.protobuf import service
from google.protobuf import service_reflection
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)



DESCRIPTOR = descriptor.FileDescriptor(
  name='events.proto',
  package='leap.common.events',
  serialized_pb='\n\x0c\x65vents.proto\x12\x12leap.common.events\"\x97\x01\n\rSignalRequest\x12(\n\x05\x65vent\x18\x01 \x02(\x0e\x32\x19.leap.common.events.Event\x12\x0f\n\x07\x63ontent\x18\x02 \x02(\t\x12\x12\n\nmac_method\x18\x03 \x02(\t\x12\x0b\n\x03mac\x18\x04 \x02(\x0c\x12\x12\n\nenc_method\x18\x05 \x01(\t\x12\x16\n\x0e\x65rror_occurred\x18\x06 \x01(\x08\"j\n\x0fRegisterRequest\x12(\n\x05\x65vent\x18\x01 \x02(\x0e\x32\x19.leap.common.events.Event\x12\x0c\n\x04port\x18\x02 \x02(\x05\x12\x12\n\nmac_method\x18\x03 \x02(\t\x12\x0b\n\x03mac\x18\x04 \x02(\x0c\"l\n\x11UnregisterRequest\x12(\n\x05\x65vent\x18\x01 \x02(\x0e\x32\x19.leap.common.events.Event\x12\x0c\n\x04port\x18\x02 \x02(\x05\x12\x12\n\nmac_method\x18\x03 \x02(\t\x12\x0b\n\x03mac\x18\x04 \x02(\x0c\"\r\n\x0bPingRequest\"\x82\x01\n\rEventResponse\x12\x38\n\x06status\x18\x01 \x02(\x0e\x32(.leap.common.events.EventResponse.Status\x12\x0e\n\x06result\x18\x02 \x01(\t\"\'\n\x06Status\x12\x06\n\x02OK\x10\x01\x12\n\n\x06UNAUTH\x10\x02\x12\t\n\x05\x45RROR\x10\x03*\x9f\n\n\x05\x45vent\x12\x15\n\x11\x43LIENT_SESSION_ID\x10\x01\x12\x0e\n\nCLIENT_UID\x10\x02\x12\x19\n\x15SOLEDAD_CREATING_KEYS\x10\x03\x12\x1e\n\x1aSOLEDAD_DONE_CREATING_KEYS\x10\x04\x12\x1a\n\x16SOLEDAD_UPLOADING_KEYS\x10\x05\x12\x1f\n\x1bSOLEDAD_DONE_UPLOADING_KEYS\x10\x06\x12\x1c\n\x18SOLEDAD_DOWNLOADING_KEYS\x10\x07\x12!\n\x1dSOLEDAD_DONE_DOWNLOADING_KEYS\x10\x08\x12\x1c\n\x18SOLEDAD_NEW_DATA_TO_SYNC\x10\t\x12\x1a\n\x16SOLEDAD_DONE_DATA_SYNC\x10\n\x12\x17\n\x13UPDATER_NEW_UPDATES\x10\x0b\x12\x19\n\x15UPDATER_DONE_UPDATING\x10\x0c\x12\x10\n\x0cRAISE_WINDOW\x10\r\x12\x18\n\x14SMTP_SERVICE_STARTED\x10\x0e\x12 \n\x1cSMTP_SERVICE_FAILED_TO_START\x10\x0f\x12%\n!SMTP_RECIPIENT_ACCEPTED_ENCRYPTED\x10\x10\x12\'\n#SMTP_RECIPIENT_ACCEPTED_UNENCRYPTED\x10\x11\x12\x1b\n\x17SMTP_RECIPIENT_REJECTED\x10\x12\x12\x1f\n\x1bSMTP_START_ENCRYPT_AND_SIGN\x10\x13\x12\x1d\n\x19SMTP_END_ENCRYPT_AND_SIGN\x10\x14\x12\x13\n\x0fSMTP_START_SIGN\x10\x15\x12\x11\n\rSMTP_END_SIGN\x10\x16\x12\x1b\n\x17SMTP_SEND_MESSAGE_START\x10\x17\x12\x1d\n\x19SMTP_SEND_MESSAGE_SUCCESS\x10\x18\x12\x1b\n\x17SMTP_SEND_MESSAGE_ERROR\x10\x19\x12\x18\n\x14SMTP_CONNECTION_LOST\x10\x1a\x12\x18\n\x14IMAP_SERVICE_STARTED\x10\x1e\x12 \n\x1cIMAP_SERVICE_FAILED_TO_START\x10\x1f\x12\x15\n\x11IMAP_CLIENT_LOGIN\x10 \x12\x19\n\x15IMAP_FETCHED_INCOMING\x10!\x12\x17\n\x13IMAP_MSG_PROCESSING\x10\"\x12\x16\n\x12IMAP_MSG_DECRYPTED\x10#\x12\x1a\n\x16IMAP_MSG_SAVED_LOCALLY\x10$\x12\x1d\n\x19IMAP_MSG_DELETED_INCOMING\x10%\x12\x18\n\x14IMAP_UNHANDLED_ERROR\x10&\x12\x14\n\x10IMAP_UNREAD_MAIL\x10\'\x12\x1e\n\x1aKEYMANAGER_LOOKING_FOR_KEY\x10(\x12\x18\n\x14KEYMANAGER_KEY_FOUND\x10)\x12\x1c\n\x18KEYMANAGER_KEY_NOT_FOUND\x10*\x12%\n!KEYMANAGER_STARTED_KEY_GENERATION\x10+\x12&\n\"KEYMANAGER_FINISHED_KEY_GENERATION\x10,\x12\"\n\x1eKEYMANAGER_DONE_UPLOADING_KEYS\x10-\x12\x1e\n\x1aSOLEDAD_INVALID_AUTH_TOKEN\x10.\x12\x1c\n\x18SOLEDAD_SYNC_SEND_STATUS\x10/\x12\x1f\n\x1bSOLEDAD_SYNC_RECEIVE_STATUS\x10\x30\x32\xdd\x02\n\x13\x45ventsServerService\x12J\n\x04ping\x12\x1f.leap.common.events.PingRequest\x1a!.leap.common.events.EventResponse\x12R\n\x08register\x12#.leap.common.events.RegisterRequest\x1a!.leap.common.events.EventResponse\x12V\n\nunregister\x12%.leap.common.events.UnregisterRequest\x1a!.leap.common.events.EventResponse\x12N\n\x06signal\x12!.leap.common.events.SignalRequest\x1a!.leap.common.events.EventResponse2\xb1\x01\n\x13\x45ventsClientService\x12J\n\x04ping\x12\x1f.leap.common.events.PingRequest\x1a!.leap.common.events.EventResponse\x12N\n\x06signal\x12!.leap.common.events.SignalRequest\x1a!.leap.common.events.EventResponseB\x03\x90\x01\x01')

_EVENT = descriptor.EnumDescriptor(
  name='Event',
  full_name='leap.common.events.Event',
  filename=None,
  file=DESCRIPTOR,
  values=[
    descriptor.EnumValueDescriptor(
      name='CLIENT_SESSION_ID', index=0, number=1,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='CLIENT_UID', index=1, number=2,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SOLEDAD_CREATING_KEYS', index=2, number=3,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SOLEDAD_DONE_CREATING_KEYS', index=3, number=4,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SOLEDAD_UPLOADING_KEYS', index=4, number=5,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SOLEDAD_DONE_UPLOADING_KEYS', index=5, number=6,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SOLEDAD_DOWNLOADING_KEYS', index=6, number=7,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SOLEDAD_DONE_DOWNLOADING_KEYS', index=7, number=8,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SOLEDAD_NEW_DATA_TO_SYNC', index=8, number=9,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SOLEDAD_DONE_DATA_SYNC', index=9, number=10,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='UPDATER_NEW_UPDATES', index=10, number=11,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='UPDATER_DONE_UPDATING', index=11, number=12,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='RAISE_WINDOW', index=12, number=13,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SMTP_SERVICE_STARTED', index=13, number=14,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SMTP_SERVICE_FAILED_TO_START', index=14, number=15,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SMTP_RECIPIENT_ACCEPTED_ENCRYPTED', index=15, number=16,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SMTP_RECIPIENT_ACCEPTED_UNENCRYPTED', index=16, number=17,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SMTP_RECIPIENT_REJECTED', index=17, number=18,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SMTP_START_ENCRYPT_AND_SIGN', index=18, number=19,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SMTP_END_ENCRYPT_AND_SIGN', index=19, number=20,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SMTP_START_SIGN', index=20, number=21,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SMTP_END_SIGN', index=21, number=22,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SMTP_SEND_MESSAGE_START', index=22, number=23,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SMTP_SEND_MESSAGE_SUCCESS', index=23, number=24,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SMTP_SEND_MESSAGE_ERROR', index=24, number=25,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SMTP_CONNECTION_LOST', index=25, number=26,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='IMAP_SERVICE_STARTED', index=26, number=30,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='IMAP_SERVICE_FAILED_TO_START', index=27, number=31,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='IMAP_CLIENT_LOGIN', index=28, number=32,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='IMAP_FETCHED_INCOMING', index=29, number=33,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='IMAP_MSG_PROCESSING', index=30, number=34,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='IMAP_MSG_DECRYPTED', index=31, number=35,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='IMAP_MSG_SAVED_LOCALLY', index=32, number=36,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='IMAP_MSG_DELETED_INCOMING', index=33, number=37,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='IMAP_UNHANDLED_ERROR', index=34, number=38,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='IMAP_UNREAD_MAIL', index=35, number=39,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='KEYMANAGER_LOOKING_FOR_KEY', index=36, number=40,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='KEYMANAGER_KEY_FOUND', index=37, number=41,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='KEYMANAGER_KEY_NOT_FOUND', index=38, number=42,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='KEYMANAGER_STARTED_KEY_GENERATION', index=39, number=43,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='KEYMANAGER_FINISHED_KEY_GENERATION', index=40, number=44,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='KEYMANAGER_DONE_UPLOADING_KEYS', index=41, number=45,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SOLEDAD_INVALID_AUTH_TOKEN', index=42, number=46,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SOLEDAD_SYNC_SEND_STATUS', index=43, number=47,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='SOLEDAD_SYNC_RECEIVE_STATUS', index=44, number=48,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=557,
  serialized_end=1868,
)


CLIENT_SESSION_ID = 1
CLIENT_UID = 2
SOLEDAD_CREATING_KEYS = 3
SOLEDAD_DONE_CREATING_KEYS = 4
SOLEDAD_UPLOADING_KEYS = 5
SOLEDAD_DONE_UPLOADING_KEYS = 6
SOLEDAD_DOWNLOADING_KEYS = 7
SOLEDAD_DONE_DOWNLOADING_KEYS = 8
SOLEDAD_NEW_DATA_TO_SYNC = 9
SOLEDAD_DONE_DATA_SYNC = 10
UPDATER_NEW_UPDATES = 11
UPDATER_DONE_UPDATING = 12
RAISE_WINDOW = 13
SMTP_SERVICE_STARTED = 14
SMTP_SERVICE_FAILED_TO_START = 15
SMTP_RECIPIENT_ACCEPTED_ENCRYPTED = 16
SMTP_RECIPIENT_ACCEPTED_UNENCRYPTED = 17
SMTP_RECIPIENT_REJECTED = 18
SMTP_START_ENCRYPT_AND_SIGN = 19
SMTP_END_ENCRYPT_AND_SIGN = 20
SMTP_START_SIGN = 21
SMTP_END_SIGN = 22
SMTP_SEND_MESSAGE_START = 23
SMTP_SEND_MESSAGE_SUCCESS = 24
SMTP_SEND_MESSAGE_ERROR = 25
SMTP_CONNECTION_LOST = 26
IMAP_SERVICE_STARTED = 30
IMAP_SERVICE_FAILED_TO_START = 31
IMAP_CLIENT_LOGIN = 32
IMAP_FETCHED_INCOMING = 33
IMAP_MSG_PROCESSING = 34
IMAP_MSG_DECRYPTED = 35
IMAP_MSG_SAVED_LOCALLY = 36
IMAP_MSG_DELETED_INCOMING = 37
IMAP_UNHANDLED_ERROR = 38
IMAP_UNREAD_MAIL = 39
KEYMANAGER_LOOKING_FOR_KEY = 40
KEYMANAGER_KEY_FOUND = 41
KEYMANAGER_KEY_NOT_FOUND = 42
KEYMANAGER_STARTED_KEY_GENERATION = 43
KEYMANAGER_FINISHED_KEY_GENERATION = 44
KEYMANAGER_DONE_UPLOADING_KEYS = 45
SOLEDAD_INVALID_AUTH_TOKEN = 46
SOLEDAD_SYNC_SEND_STATUS = 47
SOLEDAD_SYNC_RECEIVE_STATUS = 48


_EVENTRESPONSE_STATUS = descriptor.EnumDescriptor(
  name='Status',
  full_name='leap.common.events.EventResponse.Status',
  filename=None,
  file=DESCRIPTOR,
  values=[
    descriptor.EnumValueDescriptor(
      name='OK', index=0, number=1,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='UNAUTH', index=1, number=2,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='ERROR', index=2, number=3,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=515,
  serialized_end=554,
)


_SIGNALREQUEST = descriptor.Descriptor(
  name='SignalRequest',
  full_name='leap.common.events.SignalRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='event', full_name='leap.common.events.SignalRequest.event', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='content', full_name='leap.common.events.SignalRequest.content', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='mac_method', full_name='leap.common.events.SignalRequest.mac_method', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='mac', full_name='leap.common.events.SignalRequest.mac', index=3,
      number=4, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='enc_method', full_name='leap.common.events.SignalRequest.enc_method', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='error_occurred', full_name='leap.common.events.SignalRequest.error_occurred', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=37,
  serialized_end=188,
)


_REGISTERREQUEST = descriptor.Descriptor(
  name='RegisterRequest',
  full_name='leap.common.events.RegisterRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='event', full_name='leap.common.events.RegisterRequest.event', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='port', full_name='leap.common.events.RegisterRequest.port', index=1,
      number=2, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='mac_method', full_name='leap.common.events.RegisterRequest.mac_method', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='mac', full_name='leap.common.events.RegisterRequest.mac', index=3,
      number=4, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=190,
  serialized_end=296,
)


_UNREGISTERREQUEST = descriptor.Descriptor(
  name='UnregisterRequest',
  full_name='leap.common.events.UnregisterRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='event', full_name='leap.common.events.UnregisterRequest.event', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='port', full_name='leap.common.events.UnregisterRequest.port', index=1,
      number=2, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='mac_method', full_name='leap.common.events.UnregisterRequest.mac_method', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='mac', full_name='leap.common.events.UnregisterRequest.mac', index=3,
      number=4, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=298,
  serialized_end=406,
)


_PINGREQUEST = descriptor.Descriptor(
  name='PingRequest',
  full_name='leap.common.events.PingRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=408,
  serialized_end=421,
)


_EVENTRESPONSE = descriptor.Descriptor(
  name='EventResponse',
  full_name='leap.common.events.EventResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='status', full_name='leap.common.events.EventResponse.status', index=0,
      number=1, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='result', full_name='leap.common.events.EventResponse.result', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _EVENTRESPONSE_STATUS,
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=424,
  serialized_end=554,
)

_SIGNALREQUEST.fields_by_name['event'].enum_type = _EVENT
_REGISTERREQUEST.fields_by_name['event'].enum_type = _EVENT
_UNREGISTERREQUEST.fields_by_name['event'].enum_type = _EVENT
_EVENTRESPONSE.fields_by_name['status'].enum_type = _EVENTRESPONSE_STATUS
_EVENTRESPONSE_STATUS.containing_type = _EVENTRESPONSE;
DESCRIPTOR.message_types_by_name['SignalRequest'] = _SIGNALREQUEST
DESCRIPTOR.message_types_by_name['RegisterRequest'] = _REGISTERREQUEST
DESCRIPTOR.message_types_by_name['UnregisterRequest'] = _UNREGISTERREQUEST
DESCRIPTOR.message_types_by_name['PingRequest'] = _PINGREQUEST
DESCRIPTOR.message_types_by_name['EventResponse'] = _EVENTRESPONSE

class SignalRequest(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _SIGNALREQUEST
  
  # @@protoc_insertion_point(class_scope:leap.common.events.SignalRequest)

class RegisterRequest(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _REGISTERREQUEST
  
  # @@protoc_insertion_point(class_scope:leap.common.events.RegisterRequest)

class UnregisterRequest(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _UNREGISTERREQUEST
  
  # @@protoc_insertion_point(class_scope:leap.common.events.UnregisterRequest)

class PingRequest(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _PINGREQUEST
  
  # @@protoc_insertion_point(class_scope:leap.common.events.PingRequest)

class EventResponse(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _EVENTRESPONSE
  
  # @@protoc_insertion_point(class_scope:leap.common.events.EventResponse)


_EVENTSSERVERSERVICE = descriptor.ServiceDescriptor(
  name='EventsServerService',
  full_name='leap.common.events.EventsServerService',
  file=DESCRIPTOR,
  index=0,
  options=None,
  serialized_start=1871,
  serialized_end=2220,
  methods=[
  descriptor.MethodDescriptor(
    name='ping',
    full_name='leap.common.events.EventsServerService.ping',
    index=0,
    containing_service=None,
    input_type=_PINGREQUEST,
    output_type=_EVENTRESPONSE,
    options=None,
  ),
  descriptor.MethodDescriptor(
    name='register',
    full_name='leap.common.events.EventsServerService.register',
    index=1,
    containing_service=None,
    input_type=_REGISTERREQUEST,
    output_type=_EVENTRESPONSE,
    options=None,
  ),
  descriptor.MethodDescriptor(
    name='unregister',
    full_name='leap.common.events.EventsServerService.unregister',
    index=2,
    containing_service=None,
    input_type=_UNREGISTERREQUEST,
    output_type=_EVENTRESPONSE,
    options=None,
  ),
  descriptor.MethodDescriptor(
    name='signal',
    full_name='leap.common.events.EventsServerService.signal',
    index=3,
    containing_service=None,
    input_type=_SIGNALREQUEST,
    output_type=_EVENTRESPONSE,
    options=None,
  ),
])

class EventsServerService(service.Service):
  __metaclass__ = service_reflection.GeneratedServiceType
  DESCRIPTOR = _EVENTSSERVERSERVICE
class EventsServerService_Stub(EventsServerService):
  __metaclass__ = service_reflection.GeneratedServiceStubType
  DESCRIPTOR = _EVENTSSERVERSERVICE


_EVENTSCLIENTSERVICE = descriptor.ServiceDescriptor(
  name='EventsClientService',
  full_name='leap.common.events.EventsClientService',
  file=DESCRIPTOR,
  index=1,
  options=None,
  serialized_start=2223,
  serialized_end=2400,
  methods=[
  descriptor.MethodDescriptor(
    name='ping',
    full_name='leap.common.events.EventsClientService.ping',
    index=0,
    containing_service=None,
    input_type=_PINGREQUEST,
    output_type=_EVENTRESPONSE,
    options=None,
  ),
  descriptor.MethodDescriptor(
    name='signal',
    full_name='leap.common.events.EventsClientService.signal',
    index=1,
    containing_service=None,
    input_type=_SIGNALREQUEST,
    output_type=_EVENTRESPONSE,
    options=None,
  ),
])

class EventsClientService(service.Service):
  __metaclass__ = service_reflection.GeneratedServiceType
  DESCRIPTOR = _EVENTSCLIENTSERVICE
class EventsClientService_Stub(EventsClientService):
  __metaclass__ = service_reflection.GeneratedServiceStubType
  DESCRIPTOR = _EVENTSCLIENTSERVICE

# @@protoc_insertion_point(module_scope)
